class ClassRegistry(object):
    """
    A registry to store classes identified by a unique name.

    :param attr_name: On each registered class, the unique name will be saved
                      under this class attribute name.
    """

    def __init__(self, attr_name="name"):
        self._classes = {}
        self.attr_name = attr_name

    def register(self, name, attr_value=None, **kwargs):
        """
        Register a class with a given name.

        :param name: name to identify the class
        :param attr_value: If given, it will be used to define the value of
                           the class attribute. If not given, *name* is used.
        :param kwargs: key and values to add to the class object
        """
        assert name not in self._classes

        def wrapper(klass):
            self._classes[name] = klass
            setattr(klass, self.attr_name, attr_value or name)
            for key, value in kwargs.items():
                setattr(klass, key, value)
            return klass

        return wrapper

    def get(self, name):
        """
        Returns a registered class.
        """
        return self._classes[name]

    def names(self, predicate=None):
        """
        Returns a list of registered class names.

        :param predicate: if provided, should be a function to filter
                          classes. Only the names of the classes where
                          predicate returns True will be included.
        """
        names = sorted(self._classes)
        if predicate:
            for name, klass in self._classes.items():
                if not predicate(klass):
                    names.remove(name)
        return names



"""
Reading and writing of the configuration file.
"""

from __future__ import absolute_import, print_function

import os

import mozinfo
from configobj import ConfigObj, ParseError

from mozregression.errors import MozRegressionError
from mozregression.log import colorize

CONFIG_FILE_HELP_URL = "http://mozilla.github.io/mozregression/documentation/configuration.html"
DEFAULT_CONF_FNAME = os.path.expanduser(
    os.path.join("~", ".mozilla", "mozregression", "mozregression.cfg")
)
TC_CREDENTIALS_FNAME = os.path.expanduser(
    os.path.join("~", ".mozilla", "mozregression", "taskcluster-credentials.json")
)
ARCHIVE_BASE_URL = "https://archive.mozilla.org/pub"
# when a bisection range needs to be expanded, the following value is used to
# specify how many builds we try (if 20, we will try 20 before the lower limit,
# and another 20 after the higher limit)
DEFAULT_EXPAND = 20

# default values when not defined in config file.
# Note that this is also the list of options that can be used in config file
DEFAULTS = {
    "adb-profile-dir": None,
    "app": "firefox",
    "approx-policy": "auto",
    "archive-base-url": ARCHIVE_BASE_URL,
    "background-dl-policy": "cancel",
    "bits": None,
    "build-type": "",
    "cmdargs": [],
    "http-timeout": 30.0,
    "mode": "classic",
    "no-background-dl": "",
    "persist": None,
    "persist-size-limit": 0,
    "process-output": None,
    "profile": None,
    "profile-persistence": "clone",
    "repo": None,
    "taskcluster-accesstoken": None,
    "taskcluster-clientid": None,
    "enable-telemetry": True,
}


def get_config(conf_path):
    """
    Get custom defaults from configuration file in argument.
    """
    config = dict(DEFAULTS)
    try:
        config_obj = ConfigObj(conf_path)
    except ParseError as exc:
        raise MozRegressionError("Error while reading the config file %s:\n  %s" % (conf_path, exc))
    config.update(config_obj)

    return config


def _get_persist_dir(default):
    print(
        "You should configure a persist directory, where to put downloaded"
        " build files to reuse them in future bisections."
    )
    print(
        "I recommend using %s. Leave blank to use that default. If you"
        " really don't want a persist dir type NONE, else you can"
        " just define a path that you would like to use." % default
    )
    value = input("persist: ")
    if value == "NONE":
        return ""
    elif value:
        persist_dir = os.path.realpath(value)
    else:
        persist_dir = default

    if persist_dir:
        if not os.path.isdir(persist_dir):
            os.makedirs(persist_dir)
    return persist_dir


def _get_persist_size_limit(default):
    print(
        "You should choose a size limit for the persist dir. I recommend you"
        " to use %s GiB, so leave it blank to use that default. Else you"
        " can type NONE to not limit the persist dir, or any number you like"
        " (a GiB value, so type 0.5 to allow ~500 MiB)." % default
    )
    value = input("persist-size-limit: ")
    if value == "NONE":
        return 0.0
    elif value:
        return float(value)
    return default


def _get_bits(default):
    print(
        "You are using a 64-bit system, so mozregression will by default"
        " use the 64-bit build files. If you want to change that to"
        " 32-bit by default, type 32 here."
    )
    while 1:
        value = input("bits: ")
        if value in ("", "32", "64"):
            break
    if not value:
        return default
    return value


CONF_HELP = """\
# ------ mozregression configuration file ------

# Most of the command line options can be used in here.
# Just remove the -- from the long option names, e.g.

# bits = 32
# persist-size-limit = 15.0


"""


def write_config(conf_path):
    conf_dir = os.path.dirname(conf_path)
    if not os.path.isdir(conf_dir):
        os.makedirs(conf_dir)

    config = ConfigObj(conf_path, encoding="UTF8")
    if not config.initial_comment:
        config.initial_comment = CONF_HELP.splitlines()

    def _set_option(optname, getfunc, default):
        print()
        if optname not in config:
            value = getfunc(default)
            if value is not None:
                config[optname] = value
            else:
                value = default
        else:
            print("%s already defined." % optname)
            value = config[optname]
        name = colorize("{fGREEN}%s{sRESET_ALL}" % optname)
        print("%s: %s" % (name, value))

    _set_option("persist", _get_persist_dir, os.path.join(conf_dir, "persist"))

    _set_option("persist-size-limit", _get_persist_size_limit, 20.0)

    if mozinfo.os != "mac" and mozinfo.bits == 64:
        _set_option("bits", _get_bits, 64)

    config.write()

    print()
    print(colorize("Config file {sBRIGHT}%s{sRESET_ALL} written." % conf_path))
    print(
        "Note you can edit it manually, and there are other options you can"
        " configure. See %s." % CONFIG_FILE_HELP_URL
    )



from __future__ import absolute_import

import re

from mozregression.json_pushes import JsonPushes

RE_BUG_ID = re.compile(r"bug\s+(\d+)", re.I)


def find_bugids_in_push(branch, changeset):
    jp = JsonPushes(branch)
    push = jp.push(changeset, full="1")
    branches = set()
    for chset in push.changesets:
        res = RE_BUG_ID.search(chset["desc"])
        if res:
            branches.add(res.group(1))
    return [b for b in branches]


def bug_url(bugid):
    return "https://bugzilla.mozilla.org/show_bug.cgi?id=%s" % bugid



from __future__ import absolute_import

import re
from datetime import date

from mozregression.errors import UnavailableRelease
from mozregression.network import retry_get


def releases():
    """
    Provide the list of releases with their associated dates.

    The date is a string formated as "yyyy-mm-dd", and the release an integer.
    """
    # The dates comes from from https://wiki.mozilla.org/RapidRelease/Calendar,
    # using the ones in the "beta" column (formerly "aurora"). This is because
    # the merge date for beta corresponds to the last nightly for that
    # release. See bug 996812.
    releases = {
        5: "2011-04-12",
        6: "2011-05-24",
        7: "2011-07-05",
        8: "2011-08-16",
        9: "2011-09-27",
        10: "2011-11-08",
        11: "2011-12-20",
        12: "2012-01-31",
        13: "2012-03-13",
        14: "2012-04-24",
        15: "2012-06-05",
        16: "2012-07-16",
        17: "2012-08-27",
        18: "2012-10-08",
        19: "2012-11-19",
        20: "2013-01-07",
        21: "2013-02-19",
        22: "2013-04-01",
        23: "2013-05-13",
        24: "2013-06-24",
        25: "2013-08-05",
        26: "2013-09-16",
        27: "2013-10-28",
        28: "2013-12-09",
        29: "2014-02-03",
        30: "2014-03-17",
        31: "2014-04-28",
        32: "2014-06-09",
        33: "2014-07-21",
        34: "2014-09-02",
        35: "2014-10-13",
        36: "2014-11-28",
        37: "2015-01-12",
        38: "2015-02-23",
        39: "2015-03-30",
        40: "2015-05-11",
        41: "2015-06-29",
        42: "2015-08-10",
        43: "2015-09-21",
        44: "2015-10-29",
        45: "2015-12-14",
        46: "2016-01-25",
        47: "2016-03-07",
        48: "2016-04-25",
        49: "2016-06-06",
        50: "2016-08-01",
        51: "2016-09-19",
        52: "2016-11-14",
        53: "2017-01-23",
        54: "2017-03-06",
        55: "2017-06-12",
        56: "2017-08-02",
    }

    def filter_tags(tag_node):
        match = re.match(r"^FIREFOX_NIGHTLY_(\d+)_END$", tag_node["tag"])
        return int(match.group(1)) > 56 if match else False

    def map_tags(tag_node):
        release = {}
        merge_date = date.fromtimestamp(tag_node["date"][0] + tag_node["date"][1])
        ver_match = re.search(r"_(\d+)_", tag_node["tag"])
        release[int(ver_match.group(1))] = merge_date.isoformat()
        return release

    tags_url = "https://hg.mozilla.org/mozilla-central/json-tags"
    response = retry_get(tags_url)

    if response.status_code == 200:
        fetched_releases = list(map(map_tags, list(filter(filter_tags, response.json()["tags"]))))

        for release in fetched_releases:
            releases.update(release)

    return releases


def date_of_release(release):
    """
    Provide the date of a release.
    """
    try:
        return releases()[int(release)]
    except (KeyError, ValueError):
        raise UnavailableRelease(release)


def tag_of_release(release):
    """
    Provide the mercurial tag of a release, suitable for use in place of a hash
    """
    if re.match(r"^\d+$", release):
        release += ".0"
    if re.match(r"^\d+\.\d(\.\d)?$", release):
        return "FIREFOX_%s_RELEASE" % release.replace(".", "_")
    else:
        raise UnavailableRelease(release)


def tag_of_beta(release):
    """
    Provide the mercurial tag of a beta release, suitable for use in place of a
    hash
    """
    if re.match(r"^\d+\.0b\d+$", release):
        return "FIREFOX_%s_RELEASE" % release.replace(".", "_")
    elif re.match(r"^\d+(\.0)?$", release):
        return "FIREFOX_RELEASE_%s_BASE" % release.replace(".0", "")
    else:
        raise UnavailableRelease(release)


def formatted_valid_release_dates():
    """
    Returns a formatted string (ready to be printed) representing
    the valid release dates.
    """
    message = "Valid releases: \n"
    for key, value in releases().items():
        message += "% 3s: %s\n" % (key, value)

    return message



"""PyInstaller helper module to enable packaging tcl/tk with app bundles."""

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import _tkinter
import shutil
import subprocess
from pathlib import Path

import PyInstaller.log as logging
from PyInstaller.building.datastruct import Tree
from PyInstaller.building.osx import BUNDLE
from PyInstaller.compat import is_darwin
from PyInstaller.utils.hooks.tcl_tk import (
    _collect_tcl_modules,
    _find_tcl_tk,
    find_tcl_tk_shared_libs,
)

logger = logging.getLogger(__name__)


class BUNDLE_WITH_TK(BUNDLE):
    """A BUNDLE class that incorporates tcl/tk libraries."""

    def __init__(self, *args, **kwargs):
        if not is_darwin:
            # This class can only be used on macOS.
            return
        super().__init__(*args, **kwargs)

    def assemble(self):
        """Run BUNDLE.assemble, and then run post assembly methods."""
        super().assemble()
        self.post_assemble()
        self.resign_binary()

    def post_assemble(self):
        """Collect tcl/tk library/module files and add them to the bundle."""

        # These files will not be needed (originally excluded by PyInstaller).
        excluded_files = ["demos", "*.lib", "*Config.sh"]

        # Discover location of tcl/tk dynamic library files (i.e., dylib).
        _tkinter_file = _tkinter.__file__
        tcl_lib, tk_lib = find_tcl_tk_shared_libs(_tkinter_file)

        # Determine full path/location of tcl/tk libraries.
        tcl_root, tk_root = _find_tcl_tk(_tkinter_file)

        # Determine original prefixes to put all the library files in.
        tcl_prefix, tk_prefix = Path(tcl_root).name, Path(tk_root).name

        # Create Tree objects with all the files that need to be included in the bundle.
        # Tree objects are a way of creating a table of contents describing everything in
        # the provided root directory.
        tcltree = Tree(tcl_root, prefix=tcl_prefix, excludes=excluded_files)
        tktree = Tree(tk_root, prefix=tk_prefix, excludes=excluded_files)
        tclmodulestree = _collect_tcl_modules(tcl_root)
        tcl_tk_files = tcltree + tktree + tclmodulestree

        # Use Tree object to list out files that will be copied (adapted from
        # PyInstaller.building.osx).
        files = [(dest, source) for dest, source, _type in tcl_tk_files]

        # Append dynamic library files.
        files.append((tcl_lib[0], tcl_lib[1]))
        files.append((tk_lib[0], tk_lib[1]))

        # Create "lib" directory in the .app bundle.
        lib_dir = Path(self.name) / "Contents" / "lib"
        lib_dir.mkdir()

        # Iterate over all files and copy them into "lib" directory.
        for dest, source in files:
            dest_full_path = lib_dir / dest
            dest_full_path.parent.mkdir(parents=True, exist_ok=True)
            if dest_full_path.is_dir():
                shutil.copytree(source, dest_full_path)
            else:
                shutil.copy(source, dest_full_path)

    def resign_binary(self):
        """Force resigning of the .app bundle."""
        command = [
            "codesign",
            "-s",
            "-",
            "--force",
            "--all-architectures",
            "--timestamp",
            "--deep",
            "--force",
        ]
        if self.entitlements_file:
            command.append("--entitlements")
            command.append(self.entitlements_file)

        command.append(self.name)
        subprocess.check_output(command)



from setuptools import setup

# we pin these dependencies in the requirements files -- all of these
# should be python 3 compatible
DEPENDENCIES = [
    "glean_sdk>=31.1.3",
    "beautifulsoup4>=4.7.1",
    "colorama>=0.4.1",
    "configobj>=5.0.6",
    "distro>=1.8.0",
    "mozdevice>=4.1.0,<5",
    "mozfile>=2.0.0",
    "mozinfo>=1.1.0",
    "mozinstall>=2.0.0",
    "mozlog>=4.0",
    "mozprocess>=1.3.1",
    "mozprofile>=2.2.0",
    "mozrunner>=8.0.2",
    "mozversion>=2.1.0",
    "redo>=2.0.2",
    "requests>=2.21.0",
    "taskcluster>=6.0.0",
]

desc = """Regression range finder for Mozilla nightly builds"""
long_desc = """Regression range finder for Mozilla nightly builds.
For more information see the mozregression website:
http://mozilla.github.io/mozregression/"""

setup(
    name="mozregression",
    description=desc,
    long_description=long_desc,
    author="Mozilla Automation and Tools Team",
    author_email="tools@lists.mozilla.org",
    url="http://github.com/mozilla/mozregression",
    license="MPL 2.0",
    packages=["mozregression"],
    entry_points="""
          [console_scripts]
          mozregression = mozregression.main:main
        """,
    package_data={"mozregression": ["*.yaml"]},
    platforms=["Any"],
    python_requires=">=3.8",
    install_requires=DEPENDENCIES,
    classifiers=[
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
    ],
)



"""
Entry point for the mozregression command line.
"""

from __future__ import absolute_import

import atexit
import os
import pipes
import sys

import colorama
import mozfile
import requests
from mozlog import get_proxy_logger
from requests.exceptions import HTTPError, RequestException

from mozregression import __version__
from mozregression.approx_persist import ApproxPersistChooser
from mozregression.bisector import Bisection, Bisector, IntegrationHandler, NightlyHandler
from mozregression.bugzilla import bug_url, find_bugids_in_push
from mozregression.cli import cli
from mozregression.config import DEFAULT_EXPAND, TC_CREDENTIALS_FNAME
from mozregression.download_manager import BuildDownloadManager
from mozregression.errors import GoodBadExpectationError, MozRegressionError
from mozregression.fetch_build_info import IntegrationInfoFetcher, NightlyInfoFetcher
from mozregression.json_pushes import JsonPushes
from mozregression.launchers import REGISTRY as APP_REGISTRY
from mozregression.network import set_http_session
from mozregression.persist_limit import PersistLimit
from mozregression.telemetry import UsageMetrics, get_system_info, send_telemetry_ping_oop
from mozregression.tempdir import safe_mkdtemp
from mozregression.test_runner import CommandTestRunner, ManualTestRunner

LOG = get_proxy_logger("main")


class Application(object):
    def __init__(self, fetch_config, options):
        self.fetch_config = fetch_config
        self.options = options
        self._test_runner = None
        self._bisector = None
        self._build_download_manager = None
        self._download_dir = options.persist
        self._rm_download_dir = False
        if not options.persist:
            self._download_dir = safe_mkdtemp()
            self._rm_download_dir = True
        launcher_class = APP_REGISTRY.get(fetch_config.app_name)
        launcher_class.check_is_runnable()
        # init global profile if required
        self._global_profile = None
        if options.profile_persistence in ("clone-first", "reuse"):
            self._global_profile = launcher_class.create_profile(
                profile=options.profile,
                addons=options.addons,
                preferences=options.preferences,
                clone=options.profile_persistence == "clone-first",
            )
            options.cmdargs = options.cmdargs + ["--allow-downgrade"]
        elif options.profile:
            options.cmdargs = options.cmdargs + ["--allow-downgrade"]

    def clear(self):
        if self._build_download_manager:
            # cancel all possible downloads
            self._build_download_manager.cancel()
        if self._rm_download_dir:
            if self._build_download_manager:
                # we need to wait explicitly for downloading threads completion
                # here because it may remove a file in the download dir - and
                # in that case we could end up with a race condition when
                # we will remove the download dir. See
                # https://bugzilla.mozilla.org/show_bug.cgi?id=1231745
                self._build_download_manager.wait(raise_if_error=False)
            mozfile.remove(self._download_dir)
        if self._global_profile and self.options.profile_persistence == "clone-first":
            self._global_profile.cleanup()

    @property
    def test_runner(self):
        if self._test_runner is None:
            if self.options.command is None:
                self._test_runner = ManualTestRunner(
                    launcher_kwargs=dict(
                        addons=self.options.addons,
                        profile=self._global_profile or self.options.profile,
                        cmdargs=self.options.cmdargs,
                        preferences=self.options.preferences,
                        adb_profile_dir=self.options.adb_profile_dir,
                    )
                )
            else:
                self._test_runner = CommandTestRunner(self.options.command)
        return self._test_runner

    @property
    def bisector(self):
        if self._bisector is None:
            self._bisector = Bisector(
                self.fetch_config,
                self.test_runner,
                self.build_download_manager,
                dl_in_background=self.options.background_dl,
                approx_chooser=(
                    None if self.options.approx_policy != "auto" else ApproxPersistChooser(7)
                ),
            )
        return self._bisector

    @property
    def build_download_manager(self):
        if self._build_download_manager is None:
            background_dl_policy = self.options.background_dl_policy
            if not self.options.persist:
                # cancel background downloads forced
                background_dl_policy = "cancel"
            self._build_download_manager = BuildDownloadManager(
                self._download_dir,
                background_dl_policy=background_dl_policy,
                persist_limit=PersistLimit(self.options.persist_size_limit),
            )
        return self._build_download_manager

    def bisect_nightlies(self):
        good_date, bad_date = self.options.good, self.options.bad
        handler = NightlyHandler(
            find_fix=self.options.find_fix,
            ensure_good_and_bad=self.options.mode != "no-first-check",
        )
        result = self._do_bisect(handler, good_date, bad_date)
        if result == Bisection.FINISHED:
            LOG.info("Got as far as we can go bisecting nightlies...")
            handler.print_range()
            if self.fetch_config.can_go_integration():
                LOG.info("Switching bisection method to taskcluster")
                self.fetch_config.set_repo(self.fetch_config.get_nightly_repo(handler.bad_date))
                return self._bisect_integration(
                    handler.good_revision, handler.bad_revision, expand=DEFAULT_EXPAND
                )
        elif result == Bisection.USER_EXIT:
            self._print_resume_info(handler)
        else:
            # NO_DATA
            LOG.info(
                "Unable to get valid builds within the given"
                " range. You should try to launch mozregression"
                " again with a larger date range."
            )
            return 1
        return 0

    def bisect_integration(self):
        return self._bisect_integration(
            self.options.good,
            self.options.bad,
            ensure_good_and_bad=self.options.mode != "no-first-check",
        )

    def _bisect_integration(self, good_rev, bad_rev, ensure_good_and_bad=False, expand=0):
        LOG.info(
            "Getting %s builds between %s and %s"
            % (self.fetch_config.integration_branch, good_rev, bad_rev)
        )
        handler = IntegrationHandler(
            find_fix=self.options.find_fix, ensure_good_and_bad=ensure_good_and_bad
        )
        result = self._do_bisect(handler, good_rev, bad_rev, expand=expand)
        if result == Bisection.FINISHED:
            LOG.info("No more integration revisions, bisection finished.")
            handler.print_range()
            if handler.good_revision == handler.bad_revision:
                LOG.warning(
                    "It seems that you used two changesets that are in"
                    " the same push. Check the pushlog url."
                )
            elif len(handler.build_range) == 2:
                # range reduced to 2 pushes (at least ones with builds):
                # one good, one bad.
                result = handler.handle_merge()
                if result:
                    branch, good_rev, bad_rev = result
                    self.fetch_config.set_repo(branch)
                    return self._bisect_integration(good_rev, bad_rev, expand=DEFAULT_EXPAND)
                else:
                    # This code is broken, it prints out the message even when
                    # there are multiple bug numbers or commits in the range.
                    # Somebody should fix it before re-enabling it.
                    return 0
                    # print a bug if:
                    # (1) there really is only one bad push (and we're not
                    # just missing the builds for some intermediate builds)
                    # (2) there is only one bug number in that push
                    jp = JsonPushes(handler.build_range[1].repo_name)
                    num_pushes = len(
                        jp.pushes_within_changes(
                            handler.build_range[0].changeset,
                            handler.build_range[1].changeset,
                        )
                    )
                    if num_pushes == 2:
                        bugids = find_bugids_in_push(
                            handler.build_range[1].repo_name,
                            handler.build_range[1].changeset,
                        )
                        if len(bugids) == 1:
                            word = "fix" if handler.find_fix else "regression"
                            LOG.info(
                                "Looks like the following bug has the "
                                " changes which introduced the"
                                " {}:\n{}".format(word, bug_url(bugids[0]))
                            )
        elif result == Bisection.USER_EXIT:
            self._print_resume_info(handler)
        else:
            # NO_DATA. With integration branches, this can not happen if changesets
            # are incorrect - so builds are probably too old
            LOG.info(
                "There are no build artifacts for these changesets (they are probably too old)."
            )
            return 1
        return 0

    def _do_bisect(self, handler, good, bad, **kwargs):
        try:
            return self.bisector.bisect(handler, good, bad, **kwargs)
        except (KeyboardInterrupt, MozRegressionError, RequestException) as exc:
            if (
                handler.good_revision is not None
                and handler.bad_revision is not None
                and not isinstance(exc, GoodBadExpectationError)
            ):
                atexit.register(self._on_exit_print_resume_info, handler)
            raise

    def _print_resume_info(self, handler):
        # copy sys.argv, remove every --good/--bad/--repo related argument,
        # then add our own
        argv = sys.argv[:]
        args = ("--good", "--bad", "-g", "-b", "--good-rev", "--bad-rev", "--repo")
        indexes_to_remove = []
        for i, arg in enumerate(argv):
            if i in indexes_to_remove:
                continue
            for karg in args:
                if karg == arg:
                    # handle '--good 2015-01-01'
                    indexes_to_remove.extend((i, i + 1))
                    break
                elif arg.startswith(karg + "="):
                    # handle '--good=2015-01-01'
                    indexes_to_remove.append(i)
                    break
        for i in reversed(indexes_to_remove):
            del argv[i]

        argv.append("--repo=%s" % handler.build_range[0].repo_name)

        if hasattr(handler, "good_date"):
            argv.append("--good=%s" % handler.good_date)
            argv.append("--bad=%s" % handler.bad_date)
        else:
            argv.append("--good=%s" % handler.good_revision)
            argv.append("--bad=%s" % handler.bad_revision)

        LOG.info("To resume, run:")
        LOG.info(" ".join([pipes.quote(arg) for arg in argv]))

    def _on_exit_print_resume_info(self, handler):
        handler.print_range()
        self._print_resume_info(handler)

    def _launch(self, fetcher_class):
        fetcher = fetcher_class(self.fetch_config)
        build_info = fetcher.find_build_info(self.options.launch)
        self.build_download_manager.focus_download(build_info)
        self.test_runner.run_once(build_info)

    def launch_nightlies(self):
        self._launch(NightlyInfoFetcher)

    def launch_integration(self):
        self._launch(IntegrationInfoFetcher)


def pypi_latest_version():
    url = "https://pypi.python.org/pypi/mozregression/json"
    return requests.get(url, timeout=10).json()["info"]["version"]


def check_mozregression_version():
    try:
        mozregression_version = pypi_latest_version()
    except (RequestException, KeyError, ValueError):
        LOG.critical("Unable to get latest version from pypi.")
        return

    if __version__ != mozregression_version:
        LOG.warning(
            "You are using mozregression version %s, "
            "however version %s is available." % (__version__, mozregression_version)
        )

        LOG.warning(
            "You should consider upgrading via the 'pip install"
            " --upgrade mozregression' command."
        )


def main(
    argv=None,
    namespace=None,
    check_new_version=True,
    mozregression_variant="console",
):
    """
    main entry point of mozregression command line.
    """
    # terminal color support on windows
    if os.name == "nt":
        colorama.init()

    config, app = None, None
    try:
        config = cli(argv=argv, namespace=namespace)
        if check_new_version:
            check_mozregression_version()
        config.validate()
        set_http_session(get_defaults={"timeout": config.options.http_timeout})

        app = Application(config.fetch_config, config.options)
        send_telemetry_ping_oop(
            UsageMetrics(
                variant=mozregression_variant,
                appname=config.fetch_config.app_name,
                build_type=config.fetch_config.build_type,
                good=config.options.good,
                bad=config.options.bad,
                launch=config.options.launch,
                **get_system_info(),
            ),
            config.enable_telemetry,
        )

        method = getattr(app, config.action)
        sys.exit(method())

    except KeyboardInterrupt:
        sys.exit("\nInterrupted.")
    except (MozRegressionError, RequestException) as exc:
        if isinstance(exc, HTTPError) and exc.response.status_code == 401:
            # remove the taskcluster credential file - looks like it's wrong
            # anyway. This will force mozregression to ask again next time.
            mozfile.remove(TC_CREDENTIALS_FNAME)
        LOG.error(str(exc)) if config else sys.exit(str(exc))
        sys.exit(1)
    finally:
        if app:
            app.clear()


if __name__ == "__main__":
    main()
